# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "rpycocotools"
copyright = "2023, Hoel Bagard"
author = "Hoel Bagard"
release = "0.0.7"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

autodoc_typehints = "none"
master_doc = "index"

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

extensions = [
    "hoverxref.extension",
    "sphinx_codeautolink",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
]

# See https://sphinx-hoverxref.readthedocs.io/en/latest/configuration.html
hoverxref_auto_ref = True
hoverxref_domains = ["py"]
hoverxref_role_types = {
    "attr": "tooltip",
    "class": "tooltip",
    "const": "tooltip",
    "exc": "tooltip",
    "func": "tooltip",
    "meth": "tooltip",
    "mod": "tooltip",
    "obj": "tooltip",
    "ref": "tooltip",
}

intersphinx_mapping = {
    "numpy": ("https://numpy.org/doc/stable/", None),
}

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]



import pytest

import rpycocotools


@pytest.fixture()
def coco_dataset() -> rpycocotools.COCO:
    return rpycocotools.COCO("../data_samples/coco_25k/annotations.json", "../data_samples/coco_25k/images")



"""Module to benchmark how much time it takes to load a COCO dataset."""
from pathlib import Path

import pycocotools.coco
import pytest
from pytest_benchmark.fixture import BenchmarkFixture

import rpycocotools


def rpycocotools_load_dataset(annotations_path: Path | str) -> None:
    rpycocotools.COCO(str(annotations_path), "")


def pycocotools_load_dataset(annotations_path: Path | str) -> None:
    pycocotools.coco.COCO(str(annotations_path))


@pytest.mark.coco2017()
def test_rpycocotools_area_coco2017(benchmark: BenchmarkFixture, coco2017_annotations_path: str) -> None:
    benchmark.pedantic(rpycocotools_load_dataset, args=(coco2017_annotations_path, ), rounds=10, iterations=1)


@pytest.mark.coco2017()
def test_pycocotools_area_coco2017(benchmark: BenchmarkFixture, coco2017_annotations_path: str) -> None:
    benchmark.pedantic(pycocotools_load_dataset, args=(coco2017_annotations_path, ), rounds=10, iterations=1)



from hypothesis import given
from hypothesis import strategies as st

import rpycocotools
from rpycocotools.anns import Annotation, BBox, COCO_RLE

u32_max = 4_294_967_295
u32_st = st.integers(min_value=0, max_value=u32_max)


def test_access_cats(coco_dataset: rpycocotools.COCO) -> None:
    cats = coco_dataset.get_cats()
    assert len(cats) == 80


def test_access_cat(coco_dataset: rpycocotools.COCO) -> None:
    cat = coco_dataset.get_cat(2)
    assert cat.supercategory == "vehicle"
    assert cat.id == 2
    assert cat.name == "bicycle"
    assert str(cat) == "Category(id=2, name='bicycle', supercategory='vehicle')"


def test_access_anns(coco_dataset: rpycocotools.COCO) -> None:
    anns = coco_dataset.get_anns()
    assert len(anns) == 45


def test_access_ann(coco_dataset: rpycocotools.COCO) -> None:
    ann = coco_dataset.get_ann(1348739)
    assert ann.id == 1348739
    assert isinstance(ann.segmentation, rpycocotools.anns.PolygonsRS)
    assert str(ann) == ("Annotation(id=1348739, image_id=174482, category_id=3, segmentation=PolygonsRS("
                        "size=[388, 640], "
                        "counts=[[81.28, 87.23, 82.91, 83.96, 84.0, 76.33, 99.48, 76.22, 105.91, 84.5, 108.09, "
                        "93.98, 98.17, 93.44, 90.33, 94.2, 85.97, 94.53, 84.0, 94.31]]), "
                        "area=390.6123, bbox=BBox(left=81.28, top=76.22, width=26.81, height=18.31), iscrowd=0)")


def test_access_imgs(coco_dataset: rpycocotools.COCO) -> None:
    imgs = coco_dataset.get_imgs()
    assert len(imgs) == 4


def test_access_img(coco_dataset: rpycocotools.COCO) -> None:
    img = coco_dataset.get_img(img_id=174482)
    assert img.id == 174482
    assert str(img) == "Image(id=174482, width='640', height='388', file_name='000000174482.jpg')"


def test_get_img_anns(coco_dataset: rpycocotools.COCO) -> None:
    anns = coco_dataset.get_img_anns(480985)
    assert len(anns) == 13
    assert all(ann.image_id == 480985 for ann in anns)


@given(u32_st, u32_st, u32_st, u32_st)
def test_bbox_create(left: int, top: int, width: int, height: int) -> None:
    bbox = rpycocotools.anns.BBox(left, top, width, height)
    assert isinstance(bbox, rpycocotools.anns.BBox)


@given(u32_st, u32_st, u32_st, u32_st)
def test_bbox_equality(left: int, top: int, width: int, height: int) -> None:
    bbox1 = rpycocotools.anns.BBox(left, top, width, height)
    bbox2 = rpycocotools.anns.BBox(left, top, width, height)
    assert bbox1 == bbox2


@given(st.tuples(st.tuples(u32_st, u32_st, u32_st, u32_st),
                 st.tuples(u32_st, u32_st, u32_st, u32_st)).filter(lambda x: x[0] != x[1]))
def test_bbox_inequality(coords: tuple[tuple[int, int, int, int], tuple[int, int, int, int]]) -> None:
    bbox1 = rpycocotools.anns.BBox(*coords[0])
    bbox2 = rpycocotools.anns.BBox(*coords[1])
    assert bbox1 != bbox2


@given(u32_st, u32_st, u32_st, u32_st)
def test_bbox_repr(left: int, top: int, width: int, height: int) -> None:
    bbox = rpycocotools.anns.BBox(left, top, width, height)
    assert str(bbox) == f"BBox(left={left}, top={top}, width={width}, height={height})"


def test_dataset_from_components() -> None:
    expected_ann = Annotation(
        id=1,
        image_id=1,
        category_id=1,
        segmentation=COCO_RLE(size=[380, 800], counts="a"),
        area=1,
        bbox=BBox(left=1, top=1, width=1, height=1),
        iscrowd=1,
    )
    anns = [
        expected_ann,
        rpycocotools.anns.Annotation(2, 2, 1, COCO_RLE([1, 1], "b"), 1, rpycocotools.anns.BBox(1, 1, 1, 1), 1),
    ]
    imgs = [
        rpycocotools.anns.Image(1, 380, 800, "test1"),
        rpycocotools.anns.Image(2, 640, 480, "test2"),
    ]
    cats = [
        rpycocotools.anns.Category(1, "a", "c"),
        rpycocotools.anns.Category(2, "b", "c"),
    ]
    coco_dataset = rpycocotools.anns.from_dataset(imgs, anns, cats, "a")
    assert coco_dataset.get_img_anns(1)[0] == expected_ann



import pycocotools.coco
import pytest

import rpycocotools
from rpycocotools.anns import Annotation, COCO_RLE, PolygonsRS, RLE


@pytest.fixture()
def sample_annotations_path() -> str:
    return "../data_samples/coco_25k/annotations.json"


@pytest.fixture()
def coco2017_annotations_path() -> str:
    return "../data_samples/instances_train2017.json"


@pytest.fixture()
def pycocotools_sample_dataset(sample_annotations_path: str) -> pycocotools.coco.COCO:
    return pycocotools.coco.COCO(sample_annotations_path)


@pytest.fixture()
def rpycocotools_sample_dataset(sample_annotations_path: str) -> rpycocotools.COCO:
    return rpycocotools.COCO(sample_annotations_path, "")


@pytest.fixture()
def pycocotools_coco2017_dataset(coco2017_annotations_path: str) -> pycocotools.coco.COCO:
    return pycocotools.coco.COCO(coco2017_annotations_path)


@pytest.fixture()
def pycocotools_coco2017_dataset_rle(pycocotools_coco2017_dataset: pycocotools.coco.COCO) -> pycocotools.coco.COCO:
    for ann_id, ann in pycocotools_coco2017_dataset.anns.items():
        rle = pycocotools_coco2017_dataset.annToRLE(ann)
        pycocotools_coco2017_dataset.anns[ann_id]["segmentation"] = rle  # pyright: ignore[reportGeneralTypeIssues]
    for img_id, anns_list in pycocotools_coco2017_dataset.imgToAnns.items():
        rle_anns_list = []
        for ann in anns_list:
            ann["segmentation"] = pycocotools_coco2017_dataset.annToRLE(ann)  # pyright: ignore[reportGeneralTypeIssues]
            rle_anns_list.append(ann)
        pycocotools_coco2017_dataset.imgToAnns[img_id] = rle_anns_list
    return pycocotools_coco2017_dataset


@pytest.fixture()
def rpycocotools_coco2017_dataset(coco2017_annotations_path: str) -> rpycocotools.COCO:
    return rpycocotools.COCO(coco2017_annotations_path, "")


@pytest.fixture()
def rpycocotools_coco2017_dataset_rle(rpycocotools_coco2017_dataset: rpycocotools.COCO) -> rpycocotools.COCO:
    new_anns = []
    for img in rpycocotools_coco2017_dataset.get_imgs():
        for ann in rpycocotools_coco2017_dataset.get_img_anns(img.id):
            if isinstance(ann.segmentation, COCO_RLE | RLE | PolygonsRS):
                mask = rpycocotools.mask.decode(ann.segmentation)
            else:
                mask = rpycocotools.mask.decode(ann.segmentation, width=img.width, height=img.height)
            rle = rpycocotools.mask.encode(mask, target="rle")
            new_anns.append(Annotation(ann.id, ann.image_id, ann.category_id, rle, ann.area, ann.bbox, ann.iscrowd))
    return rpycocotools.anns.from_dataset(
        rpycocotools_coco2017_dataset.get_imgs(),
        new_anns,
        rpycocotools_coco2017_dataset.get_cats(),
        "",
    )



"""Module to benchmark how much time it takes to compute the total number of mask pixels in a COCO dataset."""
import pycocotools.coco
import pytest
from pytest_benchmark.fixture import BenchmarkFixture

import rpycocotools


@pytest.fixture()
def sample_annotations_total_area() -> float:
    return 293163.80235


@pytest.fixture()
def coco2017_annotations_total_area() -> float:
    return 10342252910.807093


def pycocotools_sum_dataset_masks_areas(dataset: pycocotools.coco.COCO, expected_area: float) -> None:
    total_area = 0
    for ann in dataset.anns.values():
        segmentation = dataset.annToRLE(ann)
        total_area += pycocotools.mask.area(segmentation)
    assert .99*expected_area < total_area < 1.01*expected_area


def rpycocotools_sum_dataset_masks_areas(dataset: rpycocotools.COCO, expected_area: float) -> None:
    total_area = sum([ann.area for ann in dataset.get_anns()])
    assert .99*expected_area < total_area < 1.01*expected_area


def test_rpycocotools_area(
        benchmark: BenchmarkFixture,
        rpycocotools_sample_dataset: rpycocotools.COCO,
        sample_annotations_total_area: float,
) -> None:
    benchmark.pedantic(
        rpycocotools_sum_dataset_masks_areas,
        args=(rpycocotools_sample_dataset, sample_annotations_total_area),
        rounds=500,
        iterations=10,
    )


def test_pycocotools_area(
        benchmark: BenchmarkFixture,
        pycocotools_sample_dataset: pycocotools.coco.COCO,
        sample_annotations_total_area: float,
) -> None:
    benchmark.pedantic(
        pycocotools_sum_dataset_masks_areas,
        args=(pycocotools_sample_dataset, sample_annotations_total_area),
        rounds=500,
        iterations=10,
    )


@pytest.mark.coco2017()
def test_rpycocotools_area_coco2017(
        benchmark: BenchmarkFixture,
        rpycocotools_coco2017_dataset: rpycocotools.COCO,
        coco2017_annotations_total_area: float,
) -> None:
    benchmark.pedantic(
        rpycocotools_sum_dataset_masks_areas,
        args=(rpycocotools_coco2017_dataset, coco2017_annotations_total_area),
        rounds=10,
        iterations=1,
    )


@pytest.mark.coco2017()
def test_pycocotools_area_coco2017(
        benchmark: BenchmarkFixture,
        pycocotools_coco2017_dataset: pycocotools.coco.COCO,
        coco2017_annotations_total_area: float,
) -> None:
    benchmark.pedantic(
        pycocotools_sum_dataset_masks_areas,
        args=(pycocotools_coco2017_dataset, coco2017_annotations_total_area),
        rounds=10,
        iterations=1,
    )



"""Module providing COCO annotations classes."""
from typing import Generic, TypeVar

from _rpycocotools.anns import Annotation, BBox, Category, COCO_RLE, from_dataset, Image, Polygons, PolygonsRS, RLE

_TSegmentation = TypeVar("_TSegmentation", Polygons, PolygonsRS, RLE, COCO_RLE)


class Annotation(Annotation, Generic[_TSegmentation]):
    pass



"""Module providing functions to decode and encode masks."""
from typing import Literal

import numpy as np
import numpy.typing as npt

from rpycocotools._rpycocotools import anns
from rpycocotools._rpycocotools import mask as _mask


def decode(encoded_mask: anns.RLE | anns.COCO_RLE | anns.PolygonsRS | anns.Polygons,
           width: None | int = None,
           height: None | int = None,
           ) -> npt.NDArray[np.uint8]:
    """Decode an encoded mask.

    Args:
        encoded_mask: The mask to decode. It has to be one of the 4 types of mask provided by this package.
        width: If the encoded mask of type Polygons (the format used by COCO),
               then the width of the image must be provided.
        height: If the encoded mask of type Polygons (the format used by COCO),
               then the height of the image must be provided.

    Returns:
        The decoded mask as a numpy array.
    """
    if isinstance(encoded_mask, anns.RLE):
        decoded_mask = _mask.decode_rle(encoded_mask)
    elif isinstance(encoded_mask, anns.COCO_RLE):
        decoded_mask = _mask.decode_coco_rle(encoded_mask)
    elif isinstance(encoded_mask, anns.PolygonsRS):
        decoded_mask = _mask.decode_poly_rs(encoded_mask)
    else:
        decoded_mask = _mask.decode_poly(encoded_mask, width=width, height=height)
    return decoded_mask


def encode(mask: npt.NDArray[np.uint8],
           target: Literal["rle", "coco_rle", "polygons", "polygon_rs"],
           ) -> anns.RLE | anns.COCO_RLE | anns.PolygonsRS | anns.Polygons:
    """Decode an encoded mask.

    Args:
        mask: The mask to encode, it should be a 2 dimensional array.
        target: The desired format for the encoded mask.

    Returns:
        The encoded mask.
    """
    match target:
        case "rle":
            encoded_mask = _mask.encode_to_rle(mask)
        case "coco_rle":
            encoded_mask = _mask.encode_to_coco_rle(mask)
        case "polygons":
            encoded_mask = _mask.encode_to_polygons(mask)
        case _:  # "polygons_rs"
            encoded_mask = _mask.encode_to_polygons_rs(mask)
    return encoded_mask


def area(encoded_mask: anns.RLE | anns.COCO_RLE | anns.PolygonsRS | anns.Polygons) -> int:
    """Compute the area of the given mask.

    Args:
        encoded_mask: The mask whose area should be computed.

    Returns:
        The area
    """
    if isinstance(encoded_mask, anns.RLE):
        area = _mask.area_rle(encoded_mask)
    elif isinstance(encoded_mask, anns.COCO_RLE):
        area = _mask.area_coco_rle(encoded_mask)
    elif isinstance(encoded_mask, anns.PolygonsRS):
        area = _mask.area_poly_rs(encoded_mask)
    else:
        area = _mask.area_poly(encoded_mask)
    return area


def to_bbox(encoded_mask: anns.RLE | anns.COCO_RLE | anns.PolygonsRS | anns.Polygons) -> anns.BBox:
    """Compute the bounding box of the given mask.

    Args:
        encoded_mask: The mask whose bounding box should be computed.

    Returns:
        The bounding box
    """
    if isinstance(encoded_mask, anns.RLE):
        bbox = _mask.rle_to_bbox(encoded_mask)
    elif isinstance(encoded_mask, anns.COCO_RLE):
        bbox = _mask.coco_rle_to_bbox(encoded_mask)
    elif isinstance(encoded_mask, anns.PolygonsRS):
        bbox = _mask.poly_rs_to_bbox(encoded_mask)
    else:
        bbox = _mask.poly_to_bbox(encoded_mask)
    return bbox



"""Module to benchmark how much time it takes to decode all the masks in a COCO dataset."""
import pycocotools.coco
import pytest
from pytest_benchmark.fixture import BenchmarkFixture

import rpycocotools
from rpycocotools.anns import COCO_RLE, PolygonsRS, RLE


def pycocotools_decode_masks(dataset: pycocotools.coco.COCO) -> None:
    for ann in dataset.anns.values():
        segmentation = dataset.annToRLE(ann)
        pycocotools.mask.decode(segmentation)


def rpycocotools_decode_masks(dataset: rpycocotools.COCO) -> None:
    for img in dataset.get_imgs():
        for ann in dataset.get_img_anns(img.id):
            if isinstance(ann.segmentation, COCO_RLE | RLE | PolygonsRS):
                rpycocotools.mask.decode(ann.segmentation)
            else:
                rpycocotools.mask.decode(ann.segmentation, width=img.width, height=img.height)


def test_rpycocotools_decode_sample(
        benchmark: BenchmarkFixture,
        rpycocotools_sample_dataset: rpycocotools.COCO,
) -> None:
    benchmark.pedantic(
        rpycocotools_decode_masks,
        args=(rpycocotools_sample_dataset, ),
        rounds=100,
        iterations=10,
    )


def test_pycocotools_decode_sample(
        benchmark: BenchmarkFixture,
        pycocotools_sample_dataset: pycocotools.coco.COCO,
) -> None:
    benchmark.pedantic(
        pycocotools_decode_masks,
        args=(pycocotools_sample_dataset, ),
        rounds=100,
        iterations=10,
    )


@pytest.mark.coco2017()
def test_rpycocotools_decode_coco2017(
        benchmark: BenchmarkFixture,
        rpycocotools_coco2017_dataset: rpycocotools.COCO,
) -> None:
    benchmark.pedantic(
        rpycocotools_decode_masks,
        args=(rpycocotools_coco2017_dataset, ),
        rounds=1,
        iterations=1,
    )


@pytest.mark.coco2017()
def test_rpycocotools_decode_coco2017_rle(
        benchmark: BenchmarkFixture,
        rpycocotools_coco2017_dataset_rle: rpycocotools.COCO,
) -> None:
    benchmark.pedantic(
        rpycocotools_decode_masks,
        args=(rpycocotools_coco2017_dataset_rle, ),
        rounds=1,
        iterations=1,
    )


@pytest.mark.coco2017()
def test_pycocotools_decode_coco2017(
        benchmark: BenchmarkFixture,
        pycocotools_coco2017_dataset: pycocotools.coco.COCO,
) -> None:
    benchmark.pedantic(
        pycocotools_decode_masks,
        args=(pycocotools_coco2017_dataset, ),
        rounds=1,
        iterations=1,
    )


@pytest.mark.coco2017()
def test_pycocotools_decode_coco2017_rle(
        benchmark: BenchmarkFixture,
        pycocotools_coco2017_dataset_rle: pycocotools.coco.COCO,
) -> None:
    benchmark.pedantic(
        pycocotools_decode_masks,
        args=(pycocotools_coco2017_dataset_rle, ),
        rounds=1,
        iterations=1,
    )
